@@ -33,7 +33,7 @@
 #include <string.h>
 #include <time.h>
 
-#include <glib/gi18n.h>
+#include "gtksourceview-i18n.h"
 
 #include "gtksourceview-marshal.h"
 #include "gtksourcetag.h"
@@ -378,7 +378,7 @@ gtk_source_print_job_class_init (GtkSourcePrintJobClass *klass)
 			  G_SIGNAL_RUN_LAST,
 			  G_STRUCT_OFFSET (GtkSourcePrintJobClass, begin_page),
 			  NULL, NULL,
-			  gtksourceview_marshal_VOID__VOID,
+			  _gtksourceview_marshal_VOID__VOID,
 			  G_TYPE_NONE, 
 			  0);
 	print_job_signals [FINISHED] =
@@ -387,7 +387,7 @@ gtk_source_print_job_class_init (GtkSourcePrintJobClass *klass)
 			  G_SIGNAL_RUN_FIRST,
 			  G_STRUCT_OFFSET (GtkSourcePrintJobClass, finished),
 			  NULL, NULL,
-			  gtksourceview_marshal_VOID__VOID,
+			  _gtksourceview_marshal_VOID__VOID,
 			  G_TYPE_NONE, 
 			  0);
 }
@@ -774,13 +774,17 @@ update_page_size_and_margins (GtkSourcePrintJob *job)
 						   &job->priv->page_width,
 						   &job->priv->page_height);
 
-	gnome_print_config_get_length (job->priv->config, GNOME_PRINT_KEY_PAGE_MARGIN_TOP,
+	gnome_print_config_get_length (job->priv->config,
+				       (guchar*)GNOME_PRINT_KEY_PAGE_MARGIN_TOP,
 				       &job->priv->doc_margin_top, NULL);
-	gnome_print_config_get_length (job->priv->config, GNOME_PRINT_KEY_PAGE_MARGIN_BOTTOM,
+	gnome_print_config_get_length (job->priv->config,
+				       (guchar*)GNOME_PRINT_KEY_PAGE_MARGIN_BOTTOM,
 				       &job->priv->doc_margin_bottom, NULL);
-	gnome_print_config_get_length (job->priv->config, GNOME_PRINT_KEY_PAGE_MARGIN_LEFT,
+	gnome_print_config_get_length (job->priv->config,
+				       (guchar*)GNOME_PRINT_KEY_PAGE_MARGIN_LEFT,
 				       &job->priv->doc_margin_left, NULL);
-	gnome_print_config_get_length (job->priv->config, GNOME_PRINT_KEY_PAGE_MARGIN_RIGHT,
+	gnome_print_config_get_length (job->priv->config,
+				       (guchar*)GNOME_PRINT_KEY_PAGE_MARGIN_RIGHT,
 				       &job->priv->doc_margin_right, NULL);
 
 	/* set default fonts for numbers and header/footer */
@@ -920,7 +924,7 @@ font_description_to_gnome_font_name (PangoFontDescription *desc)
  * gnome_font_face_find_closest() (probably a gnome-print bug)
  */
 static void
-face_and_size_from_full_name (const guchar   *name,
+face_and_size_from_full_name (const gchar   *name,
 			      GnomeFontFace **face,
 			      gdouble        *size)
 {
@@ -940,7 +944,7 @@ face_and_size_from_full_name (const guchar   *name,
 		*size = 12;
 	}
 
-	*face = gnome_font_face_find_closest (copy);
+	*face = gnome_font_face_find_closest ((guchar*)copy);
 	g_free (copy);
 }
 
@@ -960,7 +964,7 @@ font_description_from_gnome_font_name (const char *font_name)
 	style = gnome_font_face_is_italic (face) ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL;
 
 	desc = pango_font_description_new ();
-	pango_font_description_set_family (desc, gnome_font_face_get_family_name (face));
+	pango_font_description_set_family (desc, (const gchar*)gnome_font_face_get_family_name (face));
 	pango_font_description_set_weight (desc, weight);
 	pango_font_description_set_style (desc, style);
 	pango_font_description_set_size (desc, size * PANGO_SCALE);
@@ -1115,7 +1119,8 @@ get_text_with_style (GtkSourcePrintJob *job,
 	gboolean have_toggle;
 	
 	/* make sure the region to print is highlighted */
-	_gtk_source_buffer_highlight_region (job->priv->buffer, start, end, TRUE);
+	g_signal_emit_by_name (job->priv->buffer, "update_highlight",
+		start, end, TRUE);
 
 	next_toggle = *start;
 	have_toggle = gtk_text_iter_forward_to_tag_toggle (&next_toggle, NULL);
