@@ -1,4 +1,4 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*- 
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*-
  *  test-widget.c
  *
  *  Copyright (C) 2001
@@ -21,23 +21,32 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* If TEST_XML_MEM is defined the test program will try to detect memory
+ * allocated by xmlMalloc() but not freed by xmlFree() or freed by xmlFree()
+ * but not allocated by xmlMalloc(). */
+#define TEST_XML_MEM
+
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 #include <gtk/gtk.h>
+#include <gtksourceview/gtksourceview.h>
+#include <gtksourceview/gtksourcelanguage.h>
+#include <gtksourceview/gtksourcelanguagemanager.h>
+#include <gtksourceview/gtksourcestylemanager.h>
+#ifdef TEST_XML_MEM
+#include <libxml/xmlreader.h>
+#endif
+#ifdef USE_GNOME_VFS
 #include <libgnomevfs/gnome-vfs-init.h>
 #include <libgnomevfs/gnome-vfs-mime-utils.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
-#include <libgnomeprintui/gnome-print-job-preview.h>
-#include <gtksourceview/gtksourceview.h>
-#include <gtksourceview/gtksourcelanguage.h>
-#include <gtksourceview/gtksourcelanguagesmanager.h>
-#include <gtksourceview/gtksourceprintjob.h>
-
+#endif
 
 /* Global list of open windows */
 
 static GList *windows = NULL;
-
+static GtkSourceStyleScheme *style_scheme = NULL;
 
 /* Private data structures */
 
@@ -51,6 +60,8 @@ static GList *windows = NULL;
 
 static void       open_file_cb                   (GtkAction       *action,
 						  gpointer         user_data);
+static void       debug_thing_3_cb		 (GtkAction       *action,
+						  gpointer         user_data);
 
 static void       new_view_cb                    (GtkAction       *action,
 						  gpointer         user_data);
@@ -69,8 +80,6 @@ static void       insert_spaces_toggled_cb       (GtkAction       *action,
 static void       tabs_toggled_cb                (GtkAction       *action,
 						  GtkAction       *current,
 						  gpointer         user_data);
-static void       print_preview_cb               (GtkAction       *action,
-						  gpointer         user_data);
 
 static GtkWidget *create_view_window             (GtkSourceBuffer *buffer,
 						  GtkSourceView   *from);
@@ -82,17 +91,18 @@ static GtkActionEntry buffer_action_entries[] = {
 	{ "Open", GTK_STOCK_OPEN, "_Open", "<control>O",
 	  "Open a file", G_CALLBACK (open_file_cb) },
 	{ "Quit", GTK_STOCK_QUIT, "_Quit", "<control>Q",
-	  "Exit the application", G_CALLBACK (gtk_main_quit) },
+	  "Exit the application", G_CALLBACK (gtk_main_quit) }
 };
 
 static GtkActionEntry view_action_entries[] = {
 	{ "FileMenu", NULL, "_File" },
 	{ "ViewMenu", NULL, "_View" },
-	{ "PrintPreview", GTK_STOCK_PRINT, "_Print Preview", "<control>P",
-	  "Preview printing of the file", G_CALLBACK (print_preview_cb) },
 	{ "NewView", GTK_STOCK_NEW, "_New View", NULL,
 	  "Create a new view of the file", G_CALLBACK (new_view_cb) },
 	{ "TabsWidth", NULL, "_Tabs Width" },
+	{ "SmartHomeEnd", NULL, "_Smart Home/End" },
+	{ "DebugThing3", GTK_STOCK_FIND_AND_REPLACE, "Search and _Replace", "<control>R",
+	  "Search and Replace", G_CALLBACK (debug_thing_3_cb) },
 };
 
 static GtkToggleActionEntry toggle_entries[] = {
@@ -124,11 +134,24 @@ static GtkRadioActionEntry radio_entries[] = {
 	{ "TabsWidth12", NULL, "12", NULL, "Set tabulation width to 12 spaces", 12 }
 };
 
+static GtkRadioActionEntry smart_home_end_entries[] = {
+	{ "SmartHomeEndDisabled", NULL, "Disabled", NULL,
+	  "Smart Home/End disabled", GTKSOURCEVIEW_SMART_HOME_END_DISABLED },
+	{ "SmartHomeEndBefore", NULL, "Before", NULL,
+	  "Smart Home/End before", GTKSOURCEVIEW_SMART_HOME_END_BEFORE },
+	{ "SmartHomeEndAfter", NULL, "After", NULL,
+	  "Smart Home/End after", GTKSOURCEVIEW_SMART_HOME_END_AFTER },
+	{ "SmartHomeEndAlways", NULL, "Always", NULL,
+	  "Smart Home/End always", GTKSOURCEVIEW_SMART_HOME_END_ALWAYS }
+};
+
 static const gchar *view_ui_description =
 "<ui>"
 "  <menubar name=\"MainMenu\">"
 "    <menu action=\"FileMenu\">"
+"      <!--"
 "      <menuitem action=\"PrintPreview\"/>"
+"      -->"
 "    </menu>"
 "    <menu action=\"ViewMenu\">"
 "      <menuitem action=\"NewView\"/>"
@@ -148,6 +171,13 @@ static const gchar *view_ui_description =
 "        <menuitem action=\"TabsWidth10\"/>"
 "        <menuitem action=\"TabsWidth12\"/>"
 "      </menu>"
+"      <separator/>"
+"      <menu action=\"SmartHomeEnd\">"
+"        <menuitem action=\"SmartHomeEndDisabled\"/>"
+"        <menuitem action=\"SmartHomeEndBefore\"/>"
+"        <menuitem action=\"SmartHomeEndAfter\"/>"
+"        <menuitem action=\"SmartHomeEndAlways\"/>"
+"      </menu>"
 "    </menu>"
 "  </menubar>"
 "</ui>";
@@ -157,6 +187,7 @@ static const gchar *buffer_ui_description =
 "  <menubar name=\"MainMenu\">"
 "    <menu action=\"FileMenu\">"
 "      <menuitem action=\"Open\"/>"
+"      <menuitem action=\"DebugThing3\"/>"
 "      <separator/>"
 "      <menuitem action=\"Quit\"/>"
 "    </menu>"
@@ -174,113 +205,60 @@ error_dialog (GtkWindow *parent, const gchar *msg, ...)
 	va_list ap;
 	gchar *tmp;
 	GtkWidget *dialog;
-	
+
 	va_start (ap, msg);
 	tmp = g_strdup_vprintf (msg, ap);
 	va_end (ap);
-	
+
 	dialog = gtk_message_dialog_new (parent,
 					 GTK_DIALOG_DESTROY_WITH_PARENT,
 					 GTK_MESSAGE_ERROR,
 					 GTK_BUTTONS_OK,
 					 tmp);
 	g_free (tmp);
-	
+
 	gtk_dialog_run (GTK_DIALOG (dialog));
 	gtk_widget_destroy (dialog);
 }
 
-static gboolean 
-gtk_source_buffer_load_with_encoding (GtkSourceBuffer *source_buffer,
-				      const gchar     *filename,
-				      const gchar     *encoding,
-				      GError         **error)
+static gboolean
+gtk_source_buffer_load_file (GtkSourceBuffer *source_buffer,
+			     const gchar     *filename,
+			     GError         **error)
 {
-	GIOChannel *io;
 	GtkTextIter iter;
 	gchar *buffer;
-	gboolean reading;
-	
-	g_return_val_if_fail (source_buffer != NULL, FALSE);
-	g_return_val_if_fail (filename != NULL, FALSE);
+	GError *error_here = NULL;
+
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (source_buffer), FALSE);
+	g_return_val_if_fail (filename != NULL, FALSE);
 
-	*error = NULL;
-
-	io = g_io_channel_new_file (filename, "r", error);
-	if (!io)
+	if (!g_file_get_contents (filename, &buffer, NULL, &error_here))
 	{
-		error_dialog (NULL, "%s\nFile %s", (*error)->message, filename);
-		return FALSE;
-	}
-
-	if (g_io_channel_set_encoding (io, encoding, error) != G_IO_STATUS_NORMAL)
-	{
-		error_dialog (NULL, "Failed to set encoding:\n%s\n%s",
-			      filename, (*error)->message);
+		error_dialog (NULL, "%s\nFile %s", error_here->message, filename);
+		g_propagate_error (error, error_here);
 		return FALSE;
 	}
 
 	gtk_source_buffer_begin_not_undoable_action (source_buffer);
-
-	gtk_text_buffer_set_text (GTK_TEXT_BUFFER (source_buffer), "", 0);
-	buffer = g_malloc (READ_BUFFER_SIZE);
-	reading = TRUE;
-	while (reading)
-	{
-		gsize bytes_read;
-		GIOStatus status;
-		
-		status = g_io_channel_read_chars (io, buffer,
-						  READ_BUFFER_SIZE, &bytes_read,
-						  error);
-		switch (status)
-		{
-			case G_IO_STATUS_EOF:
-				reading = FALSE;
-				/* fall through */
-				
-			case G_IO_STATUS_NORMAL:
-				if (bytes_read == 0)
-				{
-					continue;
-				}
-				
-				gtk_text_buffer_get_end_iter (
-					GTK_TEXT_BUFFER (source_buffer), &iter);
-				gtk_text_buffer_insert (GTK_TEXT_BUFFER (source_buffer),
-							&iter, buffer, bytes_read);
-				break;
-				
-			case G_IO_STATUS_AGAIN:
-				continue;
-
-			case G_IO_STATUS_ERROR:
-			default:
-				error_dialog (NULL, "%s\nFile %s", (*error)->message, filename);
-
-				/* because of error in input we clear already loaded text */
-				gtk_text_buffer_set_text (GTK_TEXT_BUFFER (source_buffer), "", 0);
-				
-				reading = FALSE;
-				break;
-		}
-	}
-	g_free (buffer);
-	
+	gtk_text_buffer_set_text (GTK_TEXT_BUFFER (source_buffer), buffer, -1);
 	gtk_source_buffer_end_not_undoable_action (source_buffer);
-
-	g_io_channel_unref (io);
-
-	if (*error)
-		return FALSE;
-
 	gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (source_buffer), FALSE);
 
 	/* move cursor to the beginning */
 	gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (source_buffer), &iter);
 	gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (source_buffer), &iter);
 
+	{
+		GtkTextIter start, end;
+		char *text;
+		gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (source_buffer), &start, &end);
+		text = gtk_text_buffer_get_text (GTK_TEXT_BUFFER (source_buffer), &start, &end, TRUE);
+		g_assert (!strcmp (text, buffer));
+		g_free (text);
+	}
+
+	g_free (buffer);
 	return TRUE;
 }
 
@@ -295,23 +273,103 @@ remove_all_markers (GtkSourceBuffer *buffer)
 	while (markers)
 	{
 		GtkSourceMarker *marker = markers->data;
-		
+
 		gtk_source_buffer_delete_marker (buffer, marker);
 		markers = g_slist_delete_link (markers, markers);
 	}
 }
 
-static gboolean
-open_file (GtkSourceBuffer *buffer, const gchar *filename)
+/* Note this is wrong for several reasons, e.g. g_pattern_match is broken
+ * for glob matching. */
+static GtkSourceLanguage *
+get_language_for_filename (GtkSourceLanguageManager *manager,
+			   const gchar              *filename)
+{
+	const GSList *list;
+	gchar *filename_utf8;
+
+	filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
+	g_return_val_if_fail (filename_utf8 != NULL, NULL);
+
+	list = gtk_source_language_manager_get_available_languages (manager);
+
+	while (list != NULL)
+	{
+		GtkSourceLanguage *lang;
+		gchar **globs, **p;
+
+		lang = list->data;
+		list = list->next;
+
+		globs = gtk_source_language_get_globs (lang);
+		if (globs == NULL)
+			continue;
+
+		for (p = globs; *p != NULL; p++)
+		{
+			if (g_pattern_match_simple (*p, filename_utf8))
+			{
+				g_strfreev (globs);
+				g_free (filename_utf8);
+				return lang;
+			}
+		}
+
+		g_strfreev (globs);
+	}
+
+	g_free (filename_utf8);
+	return NULL;
+}
+
+/* Note this is wrong, because it ignores mime parent types and subtypes.
+ * It's fine to use in a simplish program like this, but is unacceptable
+ * in a serious text editor. */
+static GtkSourceLanguage *
+get_language_for_mime_type (GtkSourceLanguageManager *manager,
+			    const gchar              *mime)
+{
+	const GSList *list;
+
+	list = gtk_source_language_manager_get_available_languages (manager);
+
+	while (list != NULL)
+	{
+		GtkSourceLanguage *lang;
+		gchar **mimetypes, **p;
+
+		lang = list->data;
+		list = list->next;
+
+		mimetypes = gtk_source_language_get_mime_types (lang);
+
+		if (mimetypes == NULL)
+			continue;
+
+		for (p = mimetypes; *p != NULL; p++)
+		{
+			if (strcmp (*p, mime) == 0)
+			{
+				g_strfreev (mimetypes);
+				return lang;
+			}
+		}
+
+		g_strfreev (mimetypes);
+	}
+
+	return NULL;
+}
+
+static GtkSourceLanguage *
+get_language_for_file (GtkSourceLanguageManager *manager,
+		       const gchar              *filename)
 {
-	GtkSourceLanguagesManager *manager;
 	GtkSourceLanguage *language = NULL;
+
+#if defined(USE_GNOME_VFS)
 	gchar *mime_type;
-	GError *err = NULL;
 	gchar *uri;
-		
-	/* get the new language for the file mimetype */
-	manager = g_object_get_data (G_OBJECT (buffer), "languages-manager");
 
 	/* I hate this! */
 	if (g_path_is_absolute (filename))
@@ -321,7 +379,6 @@ open_file (GtkSourceBuffer *buffer, const gchar *filename)
 	else
 	{
 		gchar *curdir, *path;
-		
 		curdir = g_get_current_dir ();
 		path = g_strconcat (curdir, "/", filename, NULL);
 		g_free (curdir);
@@ -329,72 +386,58 @@ open_file (GtkSourceBuffer *buffer, const gchar *filename)
 		g_free (path);
 	}
 
-	mime_type = gnome_vfs_get_mime_type (uri);
+	if ((mime_type = gnome_vfs_get_mime_type (uri)))
+		language = get_language_for_mime_type (manager, mime_type);
+
+	g_free (mime_type);
 	g_free (uri);
-	if (mime_type)
+#endif
+
+	if (!language)
+		language = get_language_for_filename (manager, filename);
+
+	return language;
+}
+
+static gboolean
+open_file (GtkSourceBuffer *buffer, const gchar *filename)
+{
+	GtkSourceLanguageManager *manager;
+	GtkSourceLanguage *language = NULL;
+	gchar *freeme = NULL;
+	gboolean success = FALSE;
+
+	manager = g_object_get_data (G_OBJECT (buffer), "language-manager");
+
+	if (!g_path_is_absolute (filename))
 	{
-		language = gtk_source_languages_manager_get_language_from_mime_type (manager,
-										     mime_type);
-
-		if (language == NULL)
-		{
-			g_print ("No language found for mime type `%s'\n", mime_type);
-			g_object_set (G_OBJECT (buffer), "highlight", FALSE, NULL);
-		}
-		else
-		{	
-			g_object_set (G_OBJECT (buffer), "highlight", TRUE, NULL);
-
-			gtk_source_buffer_set_language (buffer, language);
-		}
-			
-		g_free (mime_type);
-	}
-	else
-	{
-		g_object_set (G_OBJECT (buffer), "highlight", FALSE, NULL);
-
-		g_warning ("Couldn't get mime type for file `%s'", filename);
+		gchar *curdir = g_get_current_dir ();
+		freeme = g_build_filename (curdir, filename, NULL);
+		filename = freeme;
+		g_free (curdir);
 	}
 
 	remove_all_markers (buffer);
-	gtk_source_buffer_load_with_encoding (buffer, filename, "utf-8", &err);
+
+	success = gtk_source_buffer_load_file (buffer, filename, NULL);
+
+	if (!success)
+		goto out;
+
+	language = get_language_for_file (manager, filename);
+
+	if (language == NULL)
+		g_print ("No language found for file `%s'\n", filename);
+
+	gtk_source_buffer_set_language (buffer, language);
 	g_object_set_data_full (G_OBJECT (buffer),
 				"filename", g_strdup (filename),
 				(GDestroyNotify) g_free);
 
-	if (err != NULL)
-	{
-		g_error_free (err);
-		return FALSE;
-	}
-	return TRUE;
-}
 
-
-/* Printing callbacks --------------------------------------------------------- */
-
-static void
-page_cb (GtkSourcePrintJob *job, gpointer user_data)
-{
-	g_print ("Printing %.2f%%    \r",
-		 100.0 * gtk_source_print_job_get_page (job) /
-		 gtk_source_print_job_get_page_count (job));
-}
-
-static void
-finished_cb (GtkSourcePrintJob *job, gpointer user_data)
-{
-	GnomePrintJob *gjob;
-	GtkWidget *preview;
-
-	g_print ("\n");
-	gjob = gtk_source_print_job_get_print_job (job);
-	preview = gnome_print_job_preview_new (gjob, "test-widget print preview");
- 	g_object_unref (gjob); 
- 	g_object_unref (job);
-	
-	gtk_widget_show (preview);
+out:
+	g_free (freeme);
+	return success;
 }
 
 
@@ -437,7 +480,7 @@ hl_line_toggled_cb (GtkAction *action, gpointer user_data)
 }
 
 
-static void 
+static void
 auto_indent_toggled_cb (GtkAction *action,
 			gpointer   user_data)
 {
@@ -447,7 +490,7 @@ auto_indent_toggled_cb (GtkAction *action,
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
-static void 
+static void
 insert_spaces_toggled_cb (GtkAction *action,
 			  gpointer   user_data)
 {
@@ -457,7 +500,7 @@ insert_spaces_toggled_cb (GtkAction *action,
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
-static void 
+static void
 tabs_toggled_cb (GtkAction *action,
 		 GtkAction *current,
 		 gpointer user_data)
@@ -468,13 +511,24 @@ tabs_toggled_cb (GtkAction *action,
 		gtk_radio_action_get_current_value (GTK_RADIO_ACTION (action)));
 }
 
+static void
+smart_home_end_toggled_cb (GtkAction *action,
+			   GtkAction *current,
+			   gpointer user_data)
+{
+	g_return_if_fail (GTK_IS_RADIO_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	gtk_source_view_set_smart_home_end (
+		GTK_SOURCE_VIEW (user_data),
+		gtk_radio_action_get_current_value (GTK_RADIO_ACTION (action)));
+}
+
 static void
 new_view_cb (GtkAction *action, gpointer user_data)
 {
 	GtkSourceBuffer *buffer;
 	GtkSourceView *view;
 	GtkWidget *window;
-	
+
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
 
 	view = GTK_SOURCE_VIEW (user_data);
@@ -485,66 +539,110 @@ new_view_cb (GtkAction *action, gpointer user_data)
 	gtk_widget_show (window);
 }
 
-static void
-print_preview_cb (GtkAction *action, gpointer user_data)
-{
-	GtkSourcePrintJob *job;
-	GtkSourceView *view;
-	GtkSourceBuffer *buffer;
-	GtkTextIter start, end;
-	gchar *filename;
-
-	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
-	
-	view = GTK_SOURCE_VIEW (user_data);
-	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (user_data)));
-	
-	job = gtk_source_print_job_new (NULL);
-	gtk_source_print_job_setup_from_view (job, view);
-	gtk_source_print_job_set_wrap_mode (job, GTK_WRAP_CHAR);
-	gtk_source_print_job_set_highlight (job, TRUE);
-	gtk_source_print_job_set_print_numbers (job, 5);
-
-	gtk_source_print_job_set_header_format (job,
-						"Printed on %A",
-						NULL,
-						"%F",
-						TRUE);
-
-	filename = g_object_get_data (G_OBJECT (buffer), "filename");
-	
-	gtk_source_print_job_set_footer_format (job,
-						"%T",
-						filename,
-						"Page %N/%Q",
-						TRUE);
-
-	gtk_source_print_job_set_print_header (job, TRUE);
-	gtk_source_print_job_set_print_footer (job, TRUE);
-	
-	gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (buffer), &start, &end);
-	if (gtk_source_print_job_print_range_async (job, &start, &end))
-	{
-		g_signal_connect (job, "begin_page", (GCallback) page_cb, NULL);
-		g_signal_connect (job, "finished", (GCallback) finished_cb, NULL);
-	}
-	else
-	{
-		g_warning ("Async print failed");
-	}
-}
-
 
 /* Buffer action callbacks ------------------------------------------------------------ */
 
-static void 
+static gboolean
+replace_dialog (GtkWidget *widget,
+		char     **what_p,
+		char     **replacement_p)
+{
+	GtkWidget *dialog;
+	GtkEntry *entry1, *entry2;
+	static char *what, *replacement;
+
+	if (!what)
+		what = g_strdup ("gtk");
+	if (!replacement)
+		replacement = g_strdup ("boo");
+
+	dialog = gtk_dialog_new_with_buttons ("Replace",
+					      GTK_WINDOW (gtk_widget_get_toplevel (widget)),
+					      GTK_DIALOG_MODAL,
+					      GTK_STOCK_CANCEL,
+					      GTK_RESPONSE_CANCEL,
+					      GTK_STOCK_OK,
+					      GTK_RESPONSE_OK,
+					      NULL);
+	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
+
+	entry1 = g_object_new (GTK_TYPE_ENTRY,
+			       "visible", TRUE,
+			       "text", what ? what : "",
+			       "activates-default", TRUE,
+			       NULL);
+	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
+				     GTK_WIDGET (entry1));
+	entry2 = g_object_new (GTK_TYPE_ENTRY,
+			       "visible", TRUE,
+			       "text", replacement ? replacement : "",
+			       "activates-default", TRUE,
+			       NULL);
+	gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox),
+				     GTK_WIDGET (entry2));
+
+	while (TRUE)
+	{
+		if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_OK)
+		{
+			gtk_widget_destroy (dialog);
+			return FALSE;
+		}
+
+		if (*gtk_entry_get_text (entry1))
+			break;
+	}
+
+	g_free (what);
+	*what_p = what = g_strdup (gtk_entry_get_text (entry1));
+	g_free (replacement);
+	*replacement_p = replacement = g_strdup (gtk_entry_get_text (entry2));
+
+	gtk_widget_destroy (dialog);
+	return TRUE;
+}
+
+static void
+debug_thing_3_cb (GtkAction *action,
+		  gpointer   user_data)
+{
+	GtkTextView *view = user_data;
+	GtkTextBuffer *buffer = gtk_text_view_get_buffer (view);
+	GtkTextIter iter;
+	char *what, *replacement;
+
+	if (!replace_dialog (GTK_WIDGET (view), &what, &replacement))
+		return;
+
+	gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
+
+	while (TRUE)
+	{
+		GtkTextIter match_start, match_end;
+
+		if (!gtk_text_iter_forward_search (&iter, what, 0,
+						   &match_start,
+						   &match_end,
+						   NULL))
+		{
+			break;
+		}
+
+		gtk_text_buffer_delete (buffer, &match_start, &match_end);
+		gtk_text_buffer_insert (buffer, &match_start, replacement, -1);
+		iter = match_start;
+	}
+}
+
+static void
 open_file_cb (GtkAction *action, gpointer user_data)
 {
 	GtkWidget *chooser;
 	gint response;
+	static gchar *last_dir;
 
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (user_data));
-	
+
 	chooser = gtk_file_chooser_dialog_new ("Open file...",
 					       NULL,
 					       GTK_FILE_CHOOSER_ACTION_OPEN,
@@ -552,19 +650,30 @@ open_file_cb (GtkAction *action, gpointer user_data)
 					       GTK_STOCK_OPEN, GTK_RESPONSE_OK,
 					       NULL);
 
+	if (last_dir == NULL)
+		last_dir = g_strdup (TOP_SRCDIR "/gtksourceview");
+
+	if (last_dir != NULL && g_path_is_absolute (last_dir))
+		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (chooser),
+						     last_dir);
+
 	response = gtk_dialog_run (GTK_DIALOG (chooser));
+
 	if (response == GTK_RESPONSE_OK)
 	{
 		gchar *filename;
 
 		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
+
 		if (filename != NULL)
 		{
+			g_free (last_dir);
+			last_dir = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (chooser));
 			open_file (GTK_SOURCE_BUFFER (user_data), filename);
 			g_free (filename);
 		}
 	}
-	
+
 	gtk_widget_destroy (chooser);
 }
 
@@ -581,22 +690,22 @@ update_cursor_position (GtkTextBuffer *buffer, gpointer user_data)
 	GtkLabel *pos_label;
 
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
-	
+
 	view = GTK_SOURCE_VIEW (user_data);
 	tabwidth = gtk_source_view_get_tabs_width (view);
 	pos_label = GTK_LABEL (g_object_get_data (G_OBJECT (view), "pos_label"));
-	
+
 	gtk_text_buffer_get_iter_at_mark (buffer,
 					  &iter,
 					  gtk_text_buffer_get_insert (buffer));
-	
+
 	chars = gtk_text_iter_get_offset (&iter);
 	row = gtk_text_iter_get_line (&iter) + 1;
-	
+
 	start = iter;
 	gtk_text_iter_set_line_offset (&start, 0);
 	col = 0;
-	
+
 	while (!gtk_text_iter_equal (&start, &iter))
 	{
 		if (gtk_text_iter_get_char (&start) == '\t')
@@ -605,16 +714,16 @@ update_cursor_position (GtkTextBuffer *buffer, gpointer user_data)
 		}
 		else
 			++col;
-		
+
 		gtk_text_iter_forward_char (&start);
 	}
-	
+
 	msg = g_strdup_printf ("char: %d, line: %d, column: %d", chars, row, col);
 	gtk_label_set_text (pos_label, msg);
       	g_free (msg);
 }
 
-static void 
+static void
 move_cursor_cb (GtkTextBuffer *buffer,
 		GtkTextIter   *cursoriter,
 		GtkTextMark   *mark,
@@ -642,7 +751,7 @@ window_deleted_cb (GtkWidget *widget, GdkEvent *ev, gpointer user_data)
 		GtkSourceView *view = GTK_SOURCE_VIEW (user_data);
 		GtkSourceBuffer *buffer = GTK_SOURCE_BUFFER (
 			gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
-		
+
 		windows = g_list_remove (windows, widget);
 
 		/* deinstall buffer motion signal handlers */
@@ -657,7 +766,7 @@ window_deleted_cb (GtkWidget *widget, GdkEvent *ev, gpointer user_data)
 		/* we return FALSE since we want the window destroyed */
 		return FALSE;
 	}
-	
+
 	return TRUE;
 }
 
@@ -666,15 +775,15 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 {
 	GtkSourceView *view;
 	GtkSourceBuffer *buffer;
-	
+
 	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (widget), FALSE);
-	
+
 	view = GTK_SOURCE_VIEW (widget);
 	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
-	
+
 	if (!gtk_source_view_get_show_line_markers (view))
 		return FALSE;
-	
+
 	/* check that the click was on the left gutter */
 	if (ev->window == gtk_text_view_get_window (GTK_TEXT_VIEW (view),
 						    GTK_TEXT_WINDOW_LEFT))
@@ -684,12 +793,12 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 		GSList *marker_list, *list_iter;
 		GtkSourceMarker *marker;
 		const gchar *marker_type;
-		
+
 		if (ev->button == 1)
 			marker_type = MARKER_TYPE_1;
 		else
 			marker_type = MARKER_TYPE_2;
-		
+
 		gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (view),
 						       GTK_TEXT_WINDOW_LEFT,
 						       ev->x, ev->y,
@@ -700,7 +809,7 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 					     &line_start,
 					     y_buf,
 					     NULL);
-		
+
 		line_end = line_start;
 		gtk_text_iter_forward_to_line_end (&line_end);
 
@@ -717,7 +826,7 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 		{
 			GtkSourceMarker *tmp = list_iter->data;
 			gchar *tmp_type = gtk_source_marker_get_marker_type (tmp);
-			
+
 			if (tmp_type && !strcmp (tmp_type, marker_type))
 			{
 				marker = tmp;
@@ -725,7 +834,7 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 			g_free (tmp_type);
 		}
 		g_slist_free (marker_list);
-		
+
 		if (marker)
 		{
 			/* a marker was found, so delete it */
@@ -738,7 +847,7 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 								  marker_type, &line_start);
 		}
 	}
-	
+
 	return FALSE;
 }
 
@@ -758,7 +867,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), NULL);
 	g_return_val_if_fail (from == NULL || GTK_IS_SOURCE_VIEW (from), NULL);
-	
+
 	/* window */
 	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 	gtk_container_set_border_width (GTK_CONTAINER (window), 0);
@@ -767,7 +876,10 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 
 	/* view */
 	view = gtk_source_view_new_with_buffer (buffer);
-	
+
+	if (style_scheme)
+		gtk_source_buffer_set_style_scheme (buffer, style_scheme);
+
 	g_signal_connect (buffer, "mark_set", G_CALLBACK (move_cursor_cb), view);
 	g_signal_connect (buffer, "changed", G_CALLBACK (update_cursor_position), view);
 	g_signal_connect (view, "button-press-event", G_CALLBACK (button_press_cb), NULL);
@@ -782,7 +894,10 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	gtk_action_group_add_radio_actions (action_group, radio_entries,
 					    G_N_ELEMENTS (radio_entries),
 					    -1, G_CALLBACK (tabs_toggled_cb), view);
-	
+	gtk_action_group_add_radio_actions (action_group, smart_home_end_entries,
+					    G_N_ELEMENTS (smart_home_end_entries),
+					    -1, G_CALLBACK (smart_home_end_toggled_cb), view);
+
 	ui_manager = gtk_ui_manager_new ();
 	gtk_ui_manager_insert_action_group (ui_manager, action_group, 0);
 	g_object_unref (action_group);
@@ -790,7 +905,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	/* save a reference to the ui manager in the window for later use */
 	g_object_set_data_full (G_OBJECT (window), "ui_manager",
 				ui_manager, (GDestroyNotify) g_object_unref);
-	
+
 	accel_group = gtk_ui_manager_get_accel_group (ui_manager);
 	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
 
@@ -819,7 +934,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	gtk_box_pack_start (GTK_BOX (vbox), pos_label, FALSE, FALSE, 0);
 
 	/* setup view */
-	font_desc = pango_font_description_from_string ("monospace 10");
+	font_desc = pango_font_description_from_string ("monospace");
 	if (font_desc != NULL)
 	{
 		gtk_widget_modify_font (view, font_desc);
@@ -839,11 +954,11 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		action = gtk_action_group_get_action (action_group, "ShowMarkers");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_show_line_markers (from));
-		
+
 		action = gtk_action_group_get_action (action_group, "ShowMargin");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_show_margin (from));
-		
+
 		action = gtk_action_group_get_action (action_group, "HlLine");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_highlight_current_line (from));
@@ -851,12 +966,12 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		action = gtk_action_group_get_action (action_group, "AutoIndent");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_auto_indent (from));
-		
+
 		action = gtk_action_group_get_action (action_group, "InsertSpaces");
 		gtk_toggle_action_set_active (
 			GTK_TOGGLE_ACTION (action),
 			gtk_source_view_get_insert_spaces_instead_of_tabs (from));
-		
+
 
 		tmp = g_strdup_printf ("TabsWidth%d", gtk_source_view_get_tabs_width (from));
 		action = gtk_action_group_get_action (action_group, tmp);
@@ -866,7 +981,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	}
 
 	/* add marker pixbufs */
-	error = NULL;	
+	error = NULL;
 	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-green.png", &error)))
 	{
 		gtk_source_view_set_marker_pixbuf (GTK_SOURCE_VIEW (view), MARKER_TYPE_1, pixbuf);
@@ -877,8 +992,8 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		g_message ("could not load marker 1 image.  Spurious messages might get triggered: %s",
 		error->message);
 		g_error_free (error);
-	} 
-	
+	}
+
 	error = NULL;
 	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-red.png", &error)))
 	{
@@ -889,9 +1004,9 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	{
 		g_message ("could not load marker 2 image.  Spurious messages might get triggered: %s",
 		error->message);
-		g_error_free (error);		
+		g_error_free (error);
 	}
-	
+
 	gtk_widget_show_all (vbox);
 
 	return window;
@@ -906,17 +1021,17 @@ create_main_window (GtkSourceBuffer *buffer)
 	GtkActionGroup *action_group;
 	GList *groups;
 	GError *error;
-	
+
 	window = create_view_window (buffer, NULL);
 	ui_manager = g_object_get_data (G_OBJECT (window), "ui_manager");
-	
+
 	/* buffer action group */
 	action_group = gtk_action_group_new ("BufferActions");
 	gtk_action_group_add_actions (action_group, buffer_action_entries,
 				      G_N_ELEMENTS (buffer_action_entries), buffer);
 	gtk_ui_manager_insert_action_group (ui_manager, action_group, 1);
 	g_object_unref (action_group);
-	
+
 	/* merge buffer ui */
 	error = NULL;
 	if (!gtk_ui_manager_add_ui_from_string (ui_manager, buffer_ui_description, -1, &error))
@@ -938,7 +1053,7 @@ create_main_window (GtkSourceBuffer *buffer)
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
 	action = gtk_action_group_get_action (action_group, "ShowMargin");
-	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
 
 	action = gtk_action_group_get_action (action_group, "AutoIndent");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
@@ -948,7 +1063,7 @@ create_main_window (GtkSourceBuffer *buffer)
 
 	action = gtk_action_group_get_action (action_group, "TabsWidth8");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
-	
+
 	return window;
 }
 
@@ -956,41 +1071,155 @@ create_main_window (GtkSourceBuffer *buffer)
 /* Buffer creation -------------------------------------------------------------- */
 
 static GtkSourceBuffer *
-create_source_buffer (GtkSourceLanguagesManager *manager)
+create_source_buffer (GtkSourceLanguageManager *manager)
 {
 	GtkSourceBuffer *buffer;
-	
+
 	buffer = GTK_SOURCE_BUFFER (gtk_source_buffer_new (NULL));
-	g_object_ref (manager);
-	g_object_set_data_full (G_OBJECT (buffer), "languages-manager",
-				manager, (GDestroyNotify) g_object_unref);
-	
+	g_object_set (G_OBJECT (buffer), "highlight", TRUE, NULL);
+	g_object_set_data_full (G_OBJECT (buffer), "language-manager",
+				g_object_ref (manager), (GDestroyNotify) g_object_unref);
+
 	return buffer;
 }
 
 
+/* XML memory management and verification functions ----------------------------- */
+
+#ifdef TEST_XML_MEM
+
+#define ALIGN 8
+
+/* my_free(malloc(n)) and free(my_malloc(n)) are invalid and
+ * abort on glibc */
+
+static gpointer
+my_malloc (gsize n_bytes)
+{
+	char *mem = malloc (n_bytes + ALIGN);
+	return mem ? mem + ALIGN : NULL;
+}
+
+static gpointer
+my_realloc (gpointer mem,
+	    gsize    n_bytes)
+{
+	if (mem)
+	{
+		char *new_mem = realloc ((char*) mem - ALIGN, n_bytes + ALIGN);
+		return new_mem ? new_mem + ALIGN : NULL;
+	}
+	else
+	{
+		return my_malloc (n_bytes);
+	}
+}
+
+static char *
+my_strdup (const char *s)
+{
+	if (s)
+	{
+		char *new_s = my_malloc (strlen (s) + 1);
+		strcpy (new_s, s);
+		return new_s;
+	}
+	else
+	{
+		return NULL;
+	}
+}
+
+static void
+my_free (gpointer mem)
+{
+	if (mem)
+		free ((char*) mem - ALIGN);
+}
+
+static void
+init_mem_stuff (void)
+{
+	if (1)
+	{
+		if (xmlMemSetup (my_free, my_malloc, my_realloc, my_strdup) != 0)
+			g_warning ("xmlMemSetup() failed");
+	}
+	else
+	{
+		GMemVTable mem_table = {
+			my_malloc,
+			my_realloc,
+			my_free,
+			NULL, NULL, NULL
+		};
+
+		g_mem_set_vtable (&mem_table);
+		g_slice_set_config (G_SLICE_CONFIG_ALWAYS_MALLOC, TRUE);
+	}
+}
+
+#endif /* TEST_XML_MEM */
+
+
 /* Program entry point ------------------------------------------------------------ */
 
 int
 main (int argc, char *argv[])
 {
 	GtkWidget *window;
-	GtkSourceLanguagesManager *lm;
+	GtkSourceLanguageManager *lm;
+	GtkSourceStyleManager *sm;
 	GtkSourceBuffer *buffer;
-	
-	/* initialization */
-	gtk_init (&argc, &argv);
+
+	gchar *builtin_lang_dirs[] = {TOP_SRCDIR "/gtksourceview/language-specs", NULL};
+	gchar **lang_dirs;
+	gboolean use_default_paths = FALSE;
+
+	gchar *style_scheme_id = NULL;
+	GOptionContext *context;
+
+	GOptionEntry entries[] = {
+	  { "style-scheme", 's', 0, G_OPTION_ARG_STRING, &style_scheme_id, "Style scheme name to use", "SCHEME"},
+	  { "default-paths", 'd', 0, G_OPTION_ARG_NONE, &use_default_paths, "Style scheme name to use", "SCHEME"},
+	  { NULL }
+	};
+
+	g_thread_init (NULL);
+
+#ifdef TEST_XML_MEM
+	init_mem_stuff ();
+#endif
+
+	context = g_option_context_new ("- test GtkSourceView widget");
+	g_option_context_add_main_entries (context, entries, NULL);
+	g_option_context_add_group (context, gtk_get_option_group (TRUE));
+	g_option_context_parse (context, &argc, &argv, NULL);
+
+// 	gdk_window_set_debug_updates (TRUE);
+
+#ifdef USE_GNOME_VFS
 	gnome_vfs_init ();
-	
+#endif
+
+	/* we do not use defaults so we don't need to install the library */
+	lang_dirs = use_default_paths ? NULL : builtin_lang_dirs;
+	lm = g_object_new (GTK_TYPE_SOURCE_LANGUAGE_MANAGER, "search-path", lang_dirs, NULL);
+	sm = g_object_new (GTK_TYPE_SOURCE_STYLE_MANAGER, "search-path", lang_dirs, NULL);
+
+	if (style_scheme_id != NULL)
+		style_scheme = gtk_source_style_manager_get_scheme (sm, style_scheme_id);
+	if (style_scheme == NULL)
+		style_scheme = gtk_source_style_manager_get_scheme (sm, "gvim");
+
 	/* create buffer */
-	lm = gtk_source_languages_manager_new ();
 	buffer = create_source_buffer (lm);
 	g_object_unref (lm);
 
 	if (argc > 1)
 		open_file (buffer, argv [1]);
 	else
-		open_file (buffer, "../gtksourceview/gtksourcebuffer.c");
+		open_file (buffer, TOP_SRCDIR "/gtksourceview/gtksourcebuffer.c");
 
 	/* create first window */
 	window = create_main_window (buffer);
@@ -1005,7 +1234,12 @@ main (int argc, char *argv[])
 	g_list_free (windows);
 	g_object_unref (buffer);
 
+	g_object_unref (sm);
+	g_free (style_scheme_id);
+
+#ifdef USE_GNOME_VFS
 	gnome_vfs_shutdown ();
-	
+#endif
+
 	return 0;
 }
